# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import List, Optional, Dict, Any

from dataclasses import dataclass, field
from dataclasses_json import config as dataclass_config

from ...common import (
    BaseResponse,
)


# Result Message


@dataclass
class Message(BaseResponse):
    """
    Message from the Deepgram Platform
    """

    message: str = ""


# Projects


@dataclass
class Project(BaseResponse):
    """
    Project object
    """

    project_id: str = ""
    name: str = ""


@dataclass
class ProjectsResponse(BaseResponse):
    """
    Projects Response object
    """

    projects: List[Project] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "projects" in _dict:
            _dict["projects"] = [
                Project.from_dict(projects) for projects in _dict["projects"]
            ]
        return _dict[key]


# Models


@dataclass
class STTDetails(BaseResponse):  # pylint: disable=too-many-instance-attributes
    """
    STTDetails class used to define the properties of the Speech-to-Text model response object.
    """

    name: str = ""
    canonical_name: str = ""
    architecture: str = ""
    languages: List[str] = field(default_factory=list)
    version: str = ""
    uuid: str = ""
    batch: bool = False
    streaming: bool = False
    formatted_output: bool = False

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "languages" in _dict:
            _dict["languages"] = [str(languages) for languages in _dict["languages"]]
        return _dict[key]


@dataclass
class TTSMetadata(BaseResponse):
    """
    TTSMetadata class used to define the properties for a given STT or TTS model.
    """

    accent: str = ""
    color: str = ""
    image: str = ""
    sample: str = ""
    tags: Optional[List[str]] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )


@dataclass
class TTSDetails(BaseResponse):
    """
    TTSDetails class used to define the properties of the Text-to-Speech model response object.
    """

    name: str = ""
    canonical_name: str = ""
    architecture: str = ""
    languages: List[str] = field(default_factory=list)
    version: str = ""
    uuid: str = ""
    metadata: Optional[TTSMetadata] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "metadata" in _dict:
            _dict["metadata"] = [
                TTSMetadata.from_dict(metadata) for metadata in _dict["metadata"]
            ]
        return _dict[key]


# responses


@dataclass
class ModelResponse(BaseResponse):
    """
    ModelResponse class used to define the properties of a single model.
    """

    name: str = ""
    canonical_name: str = ""
    architecture: str = ""
    language: str = ""
    version: str = ""
    uuid: str = ""
    metadata: Optional[TTSMetadata] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "metadata" in _dict:
            _dict["metadata"] = [
                TTSMetadata.from_dict(metadata) for metadata in _dict["metadata"]
            ]
        return _dict[key]


@dataclass
class ModelsResponse(BaseResponse):
    """
    ModelsResponse class used to obtain a list of models.
    """

    stt: List[STTDetails] = field(default_factory=list)
    tts: List[TTSDetails] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "stt" in _dict:
            _dict["stt"] = [STTDetails.from_dict(stt) for stt in _dict["stt"]]
        if "tts" in _dict:
            _dict["tts"] = [TTSDetails.from_dict(tts) for tts in _dict["tts"]]
        return _dict[key]


# Members


@dataclass
class Member(BaseResponse):
    """
    Member object
    """

    email: str = ""
    first_name: str = ""
    last_name: str = ""
    member_id: str = ""


@dataclass
class MembersResponse(BaseResponse):
    """
    Members Response object
    """

    members: List[Member] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "members" in _dict:
            _dict["members"] = [
                Member.from_dict(members) for members in _dict["members"]
            ]
        return _dict[key]


# Keys


@dataclass
class Key(BaseResponse):
    """
    Key object
    """

    api_key_id: str = ""
    key: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    comment: Optional[str] = ""
    created: str = ""
    scopes: List[str] = field(default_factory=list)
    expiration_date: str = field(
        default="", metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "scopes" in _dict:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        return _dict[key]


@dataclass
class KeyResponse(BaseResponse):
    """
    Key Response object
    """

    api_key: Key
    member: Member

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "api_key" in _dict:
            _dict["api_key"] = Key.from_dict(_dict["api_key"])
        if "member" in _dict:
            _dict["member"] = Member.from_dict(_dict["member"])
        return _dict[key]


@dataclass
class KeysResponse(BaseResponse):
    """
    Keys Response object
    """

    api_keys: List[KeyResponse] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "api_keys" in _dict:
            _dict["api_keys"] = [
                KeyResponse.from_dict(api_keys) for api_keys in _dict["api_keys"]
            ]
        return _dict[key]


# Scopes


@dataclass
class ScopesResponse(BaseResponse):
    """
    Scopes Response object
    """

    scopes: List[str] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "scopes" in _dict:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        return _dict[key]


# Invites


@dataclass
class Invite(BaseResponse):
    """
    Invite object
    """

    email: str = ""
    scope: str = ""


@dataclass
class InvitesResponse(BaseResponse):
    """
    Invites Response object
    """

    invites: List[Invite] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "invites" in _dict:
            _dict["invites"] = [
                Invite.from_dict(invites) for invites in _dict["invites"]
            ]
        return _dict[key]


# Usage


@dataclass
class Config(BaseResponse):  # pylint: disable=too-many-instance-attributes
    """
    Config object
    """

    language: str = ""
    model: str = ""
    punctuate: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    utterances: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    diarize: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    smart_format: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    interim_results: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    topics: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    intents: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    sentiment: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    summarize: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )


@dataclass
class STTUsageDetails(BaseResponse):  # pylint: disable=too-many-instance-attributes
    """
    Details object
    """

    config: Config
    usd: float = 0
    duration: Optional[float] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    total_audio: Optional[float] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    channels: Optional[int] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    streams: Optional[int] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    method: str = ""
    tier: Optional[str] = ""
    models: List[str] = field(default_factory=list)
    tags: Optional[List[str]] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    features: List[str] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "models" in _dict:
            _dict["models"] = [str(models) for models in _dict["models"]]
        if "tags" in _dict:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        if "features" in _dict:
            _dict["features"] = [str(features) for features in _dict["features"]]
        if "config" in _dict:
            _dict["config"] = Config.from_dict(_dict["config"])
        return _dict[key]


@dataclass
class Callback(BaseResponse):
    """
    Callback object
    """

    attempts: int = 0
    code: int = 0
    completed: str = ""


@dataclass
class TokenDetail(BaseResponse):
    """
    Token Detail object
    """

    feature: str = ""
    input: int = 0
    model: str = ""
    output: int = 0


@dataclass
class SpeechSegment(BaseResponse):
    """
    Speech Segment object
    """

    characters: int = 0
    model: str = ""
    tier: str = ""


@dataclass
class TTSUsageDetails(BaseResponse):
    """
    TTS Details object
    """

    duration: float = 0
    speech_segments: List[SpeechSegment] = field(default_factory=list)
    # pylint: disable=fixme
    # TODO: audio_metadata: None
    # pylint: enable=fixme

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "speech_segments" in _dict:
            _dict["speech_segments"] = [
                SpeechSegment.from_dict(speech_segments)
                for speech_segments in _dict["speech_segments"]
            ]
        return _dict[key]


@dataclass
class UsageResponse(BaseResponse):
    """
    UsageResponse object
    """

    details: Optional[STTUsageDetails] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    code: int = 0
    completed: str = ""
    message: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    tts_details: Optional[TTSUsageDetails] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    token_details: List[TokenDetail] = field(
        default_factory=list, metadata=dataclass_config(exclude=lambda f: f is list)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "details" in _dict:
            _dict["details"] = STTUsageDetails.from_dict(_dict["details"])
        if "tts_details" in _dict:
            _dict["tts_details"] = TTSUsageDetails.from_dict(_dict["tts_details"])
        if "token_details" in _dict:
            _dict["token_details"] = [
                TokenDetail.from_dict(token_details)
                for token_details in _dict["token_details"]
            ]
        return _dict[key]


@dataclass
class UsageRequest(BaseResponse):  # pylint: disable=too-many-instance-attributes
    """
    Usage Request object
    """

    response: UsageResponse
    project_uuid: str = ""
    request_id: str = ""
    created: str = ""
    path: str = ""
    api_key_id: str = ""
    callback: Optional[Callback] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    accessor: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "response" in _dict:
            _dict["response"] = UsageResponse.from_dict(_dict["response"])
        if "callback" in _dict:
            _dict["callback"] = Callback.from_dict(_dict["callback"])
        return _dict[key]


@dataclass
class UsageRequestsResponse(BaseResponse):
    """
    Usage Requests Response object
    """

    page: int = 0
    limit: int = 0
    requests: List[UsageRequest] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "requests" in _dict:
            _dict["requests"] = [
                UsageRequest.from_dict(requests) for requests in _dict["requests"]
            ]
        return _dict[key]


@dataclass
class STTTokens(BaseResponse):
    """
    STTTokens object
    """

    tokens_in: int = 0
    out: int = 0


@dataclass
class TTSTokens(BaseResponse):
    """
    TTSTokens object
    """

    characters: int = 0
    requests: int = 0


@dataclass
class UsageSummaryResults(BaseResponse):
    """
    Results object
    """

    tokens: STTTokens
    tts: TTSTokens
    start: str = ""
    end: str = ""
    hours: int = 0
    total_hours: int = 0
    requests: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "tokens" in _dict:
            _dict["tokens"] = STTTokens.from_dict(_dict["tokens"])
        if "tts" in _dict:
            _dict["tts"] = TTSTokens.from_dict(_dict["tts"])
        return _dict[key]


@dataclass
class Resolution(BaseResponse):
    """
    Resolution object
    """

    units: str = ""
    amount: int = 0


@dataclass
class UsageSummaryResponse(BaseResponse):
    """
    Usage Summary Response object
    """

    resolution: Resolution
    start: str = ""
    end: str = ""
    results: List[UsageSummaryResults] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "resolution" in _dict:
            _dict["resolution"] = Resolution.from_dict(_dict["resolution"])
        if "results" in _dict:
            _dict["results"] = [
                UsageSummaryResults.from_dict(results) for results in _dict["results"]
            ]
        return _dict[key]


@dataclass
class UsageModel(BaseResponse):
    """
    Usage Model object
    """

    name: str = ""
    language: str = ""
    version: str = ""
    model_id: str = ""


@dataclass
class UsageFieldsResponse(BaseResponse):
    """
    Usage Fields Response object
    """

    tags: Optional[List[str]] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    models: List[UsageModel] = field(default_factory=list)
    processing_methods: List[str] = field(default_factory=list)
    features: List[str] = field(default_factory=list)
    languages: Optional[List[str]] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "tags" in _dict:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        if "models" in _dict:
            _dict["models"] = [
                UsageModel.from_dict(models) for models in _dict["models"]
            ]
        if "processing_methods" in _dict:
            _dict["processing_methods"] = [
                str(processing_methods)
                for processing_methods in _dict["processing_methods"]
            ]
        if "features" in _dict:
            _dict["features"] = [str(features) for features in _dict["features"]]
        if "languages" in _dict:
            _dict["languages"] = [str(languages) for languages in _dict["languages"]]
        return _dict[key]


# Billing


@dataclass
class Balance(BaseResponse):
    """
    Balance object
    """

    balance_id: str = ""
    amount: str = ""
    units: str = ""
    purchase_order_id: str = ""


@dataclass
class BalancesResponse(BaseResponse):
    """
    Balances Response object
    """

    balances: List[Balance] = field(default_factory=list)

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "balances" in _dict:
            _dict["balances"] = [
                Balance.from_dict(balances) for balances in _dict["balances"]
            ]
        return _dict[key]



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

# rest
from .v1 import (
    #### top level
    SpeakOptions as SpeakOptionsLatest,
    SpeakRESTOptions as SpeakRESTOptionsLatest,
    # common
    TextSource as TextSourceLatest,
    BufferSource as BufferSourceLatest,
    StreamSource as StreamSourceLatest,
    FileSource as FileSourceLatest,
    # unique
    SpeakSource as SpeakSourceLatest,
    SpeakRestSource as SpeakRestSourceLatest,
    SpeakRESTSource as SpeakRESTSourceLatest,
)

from .v1 import (
    SpeakRESTClient as SpeakRESTClientLatest,
    AsyncSpeakRESTClient as AsyncSpeakRESTClientLatest,
)

from .v1 import (
    SpeakRESTResponse as SpeakRESTResponseLatest,
)

# websocket
from .v1 import (
    SpeakWebSocketClient as SpeakWebSocketClientLatest,
    AsyncSpeakWebSocketClient as AsyncSpeakWebSocketClientLatest,
    SpeakWSClient as SpeakWSClientLatest,
    AsyncSpeakWSClient as AsyncSpeakWSClientLatest,
)

from .v1 import (
    SpeakWSOptions as SpeakWSOptionsLatest,
)
from .v1 import (
    OpenResponse as OpenResponseLatest,
    SpeakWSMetadataResponse as SpeakWSMetadataResponseLatest,
    FlushedResponse as FlushedResponseLatest,
    ClearedResponse as ClearedResponseLatest,
    CloseResponse as CloseResponseLatest,
    UnhandledResponse as UnhandledResponseLatest,
    WarningResponse as WarningResponseLatest,
    ErrorResponse as ErrorResponseLatest,
)

# The client.py points to the current supported version in the SDK.
# Older versions are supported in the SDK for backwards compatibility.

# rest
# input
SpeakOptions = SpeakOptionsLatest
SpeakRESTOptions = SpeakRESTOptionsLatest
TextSource = TextSourceLatest
BufferSource = BufferSourceLatest
StreamSource = StreamSourceLatest
FileSource = FileSourceLatest
SpeakSource = SpeakSourceLatest
SpeakRestSource = SpeakRestSourceLatest
SpeakRESTSource = SpeakRESTSourceLatest  # pylint: disable=invalid-name

# output
SpeakRESTResponse = SpeakRESTResponseLatest

# websocket
# input
SpeakWSOptions = SpeakWSOptionsLatest

# output
OpenResponse = OpenResponseLatest
SpeakWSMetadataResponse = SpeakWSMetadataResponseLatest
FlushedResponse = FlushedResponseLatest
ClearedResponse = ClearedResponseLatest
CloseResponse = CloseResponseLatest
UnhandledResponse = UnhandledResponseLatest
WarningResponse = WarningResponseLatest
ErrorResponse = ErrorResponseLatest


# backward compatibility
SpeakResponse = SpeakRESTResponseLatest
SpeakClient = SpeakRESTClientLatest

# clients
SpeakRESTClient = SpeakRESTClientLatest
AsyncSpeakRESTClient = AsyncSpeakRESTClientLatest
SpeakWSClient = SpeakWSClientLatest
AsyncSpeakWSClient = AsyncSpeakWSClientLatest
SpeakWebSocketClient = SpeakWebSocketClientLatest
AsyncSpeakWebSocketClient = AsyncSpeakWebSocketClientLatest



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import ListenWebSocketClient
from .async_client import AsyncListenWebSocketClient
from .options import LiveOptions, ListenWebSocketOptions

# unique websocket response
from .response import (
    #### top level
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    #### common websocket response
    BaseResponse,
    OpenResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
    #### between rest and websocket
    ModelInfo,
    Hit,
    Search,
    #### unique
    Metadata,
    ListenWSWord,
    ListenWSAlternative,
    ListenWSChannel,
)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT


from dataclasses import dataclass

from ....common import (
    BaseResponse,
    OpenResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
)


# Speak Response Types:


@dataclass
class MetadataResponse(BaseResponse):
    """
    Metadata object
    """

    type: str = ""
    request_id: str = ""


@dataclass
class FlushedResponse(BaseResponse):
    """
    Flushed Message from the Deepgram Platform
    """

    type: str = ""
    sequence_id: int = 0


@dataclass
class ClearedResponse(BaseResponse):
    """
    Cleared object
    """

    type: str = ""
    sequence_id: int = 0


@dataclass
class WarningResponse(BaseResponse):
    """
    Warning Message from the Deepgram Platform
    """

    warn_code: str = ""
    warn_msg: str = ""
    type: str = ""



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .enums import LiveTranscriptionEvents

from .client import LiveClient
from .client import AsyncLiveClient
from .client import LiveOptions
from .client import (
    OpenResponse,
    LiveResultResponse,
    ListenWSMetadataResponse,
    MetadataResponse,  # backwards compat
    SpeechStartedResponse,
    UtteranceEndResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import logging
from typing import Dict, Union, Optional
import json

import httpx

from ....utils import verboselogs
from ....options import DeepgramClientOptions
from ...common import AbstractSyncRestClient, DeepgramError

from .response import (
    Message,
    Project,
    ProjectsResponse,
    MembersResponse,
    Key,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    Balance,
    BalancesResponse,
    ModelResponse,
    ModelsResponse,
)
from .options import (
    ProjectOptions,
    ModelOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)


class ManageClient(
    AbstractSyncRestClient
):  # pylint: disable=too-many-public-methods,too-many-lines
    """
    A client for managing Deepgram projects and associated resources via the Deepgram API.

    This class provides methods for performing various operations on Deepgram projects, including:
    - Retrieving project details
    - Updating project settings
    - Managing project members and scopes
    - Handling project invitations
    - Monitoring project usage and balances

    Args:
        config (DeepgramClientOptions): all the options for the client.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions
    _endpoint: str

    def __init__(self, config: DeepgramClientOptions):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)

        self._config = config
        self._endpoint = "v1/projects"
        super().__init__(config)

    # projects
    def list_projects(
        self,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ProjectsResponse:
        """
        List all projects for the current user.
        """
        return self.get_projects(
            timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    # pylint: disable=too-many-positional-arguments

    def get_projects(
        self,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ProjectsResponse:
        """
        Gets a list of projects for the authenticated user.

        Reference:
        https://developers.deepgram.com/reference/get-projects
        """
        self._logger.debug("ManageClient.get_projects ENTER")
        url = f"{self._config.url}/{self._endpoint}"
        self._logger.info("url: %s", url)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ProjectsResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_projects succeeded")
        self._logger.debug("ManageClient.get_projects LEAVE")
        return res

    def get_project(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Project:
        """
        Gets details for a specific project.

        Reference:
        https://developers.deepgram.com/reference/get-project
        """
        self._logger.debug("ManageClient.get_project ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Project.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_project succeeded")
        self._logger.debug("ManageClient.get_project LEAVE")
        return res

    def update_project_option(
        self,
        project_id: str,
        options: Union[Dict, ProjectOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Updates a project's settings.

        Reference:
        https://developers.deepgram.com/reference/update-project
        """
        self._logger.debug("ManageClient.update_project_option ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, ProjectOptions):
            self._logger.info("ProjectOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.patch(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("update_project_option succeeded")
        self._logger.debug("ManageClient.update_project_option LEAVE")
        return res

    def update_project(
        self,
        project_id: str,
        name: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Updates a project's settings.

        Reference:
        https://developers.deepgram.com/reference/update-project
        """
        self._logger.debug("ManageClient.update_project ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}"
        options = {
            "name": name,
        }
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.patch(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("update_project succeeded")
        self._logger.debug("ManageClient.update_project LEAVE")
        return res

    def delete_project(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Deletes a project.

        Reference:
        https://developers.deepgram.com/reference/delete-project
        """
        self._logger.debug("ManageClient.delete_project ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}"
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.delete(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("delete_project succeeded")
        self._logger.debug("ManageClient.delete_project LEAVE")
        return res

    def list_project_models(
        self,
        project_id: str,
        options: Optional[Union[Dict, ModelOptions]] = None,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelsResponse:
        """
        Please see get_project_models.
        """
        return self.get_project_models(
            project_id,
            options=options,
            timeout=timeout,
            addons=addons,
            headers=headers,
            **kwargs,
        )

    def get_project_models(
        self,
        project_id: str,
        options: Optional[Union[Dict, ModelOptions]] = None,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelsResponse:
        """
        Gets models for a specific project.

        Reference:
        https://developers.deepgram.com/reference/get-project
        https://developers.deepgram.com/reference/get-model

        Args:
            project_id (str): The ID of the project.
            options (Optional[Union[Dict, ModelOptions]]): The options for the request.
            timeout (Optional[httpx.Timeout]): The timeout setting for the request.
            addons (Optional[Dict]): Additional options for the request.
            headers (Optional[Dict]): Headers to include in the request.
            **kwargs: Additional keyword arguments.

        Returns:
            ModelsResponse: A response object containing the model details.
        """
        self._logger.debug("ManageClient.get_project_models ENTER")

        if options is None:
            options = ModelOptions()

        url = f"{self._config.url}/{self._endpoint}/{project_id}/models"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, ModelOptions):
            self._logger.info("ModelOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ModelsResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_project_models succeeded")
        self._logger.debug("ManageClient.get_project_models LEAVE")
        return res

    def get_project_model(
        self,
        project_id: str,
        model_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelResponse:
        """
        Gets a single model for a specific project.

        Reference:
        https://developers.deepgram.com/reference/get-project
        https://developers.deepgram.com/reference/get-model

        Args:
            project_id (str): The ID of the project.
            model_id (str): The ID of the model.
            timeout (Optional[httpx.Timeout]): The timeout setting for the request.
            addons (Optional[Dict]): Additional options for the request.
            headers (Optional[Dict]): Headers to include in the request.
            **kwargs: Additional keyword arguments.

        Returns:
            ModelResponse: A response object containing the model details.
        """
        self._logger.debug("ManageClient.get_project_model ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/models/{model_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("model_id: %s", model_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ModelResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_project_model succeeded")
        self._logger.debug("ManageClient.get_project_model LEAVE")
        return res

    # models
    def list_models(
        self,
        options: Optional[Union[Dict, ModelOptions]] = None,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelsResponse:
        """
        Please see get_models for more information.
        """
        return self.get_models(
            options=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    def get_models(
        self,
        options: Optional[Union[Dict, ModelOptions]] = None,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelsResponse:
        """
        Gets all models available.

        Reference:
        https://developers.deepgram.com/reference/get-model

        Args:
            options (Optional[Union[Dict, ModelOptions]]): The options for the request.
            timeout (Optional[httpx.Timeout]): The timeout setting for the request.
            addons (Optional[Dict]): Additional options for the request.
            headers (Optional[Dict]): Headers to include in the request.
            **kwargs: Additional keyword arguments.

        Returns:
            ModelsResponse: A response object containing the model details.
        """
        self._logger.debug("ManageClient.get_models ENTER")

        if options is None:
            options = ModelOptions()

        url = f"{self._config.url}/v1/models"
        self._logger.info("url: %s", url)
        if isinstance(options, ModelOptions):
            self._logger.info("ModelOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ModelsResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_models succeeded")
        self._logger.debug("ManageClient.get_models LEAVE")
        return res

    def get_model(
        self,
        model_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ModelResponse:
        """
        Gets information for a specific model.

        Reference:
        https://developers.deepgram.com/reference/get-model

        Args:
            model_id (str): The ID of the model.
            timeout (Optional[httpx.Timeout]): The timeout setting for the request.
            addons (Optional[Dict]): Additional options for the request.
            headers (Optional[Dict]): Headers to include in the request.
            **kwargs: Additional keyword arguments.

        Returns:
            ModelResponse: A response object containing the model details.
        """
        self._logger.debug("ManageClient.get_model ENTER")
        url = f"{self._config.url}/v1/models/{model_id}"
        self._logger.info("url: %s", url)
        self._logger.info("model_id: %s", model_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ModelResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_model succeeded")
        self._logger.debug("ManageClient.get_model LEAVE")
        return res

    # keys
    def list_keys(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> KeysResponse:
        """
        Please see get_keys for more information.
        """
        return self.get_keys(
            project_id, timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    def get_keys(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> KeysResponse:
        """
        Gets a list of keys for a project.

        Reference:
        https://developers.deepgram.com/reference/list-keys
        """
        self._logger.debug("ManageClient.get_keys ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/keys"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = KeysResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_keys succeeded")
        self._logger.debug("ManageClient.get_keys LEAVE")
        return res

    def get_key(
        self,
        project_id: str,
        key_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> KeyResponse:
        """
        Gets details for a specific key.

        Reference:
        https://developers.deepgram.com/reference/get-key
        """
        self._logger.debug("ManageClient.get_key ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/keys/{key_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("key_id: %s", key_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = KeyResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_key succeeded")
        self._logger.debug("ManageClient.get_key LEAVE")
        return res

    def create_key(
        self,
        project_id: str,
        options: Union[Dict, KeyOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Key:
        """
        Creates a new key.

        Reference:
        https://developers.deepgram.com/reference/create-key
        """
        self._logger.debug("ManageClient.create_key ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/keys"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, KeyOptions):
            self._logger.info("KeyOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.post(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Key.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("create_key succeeded")
        self._logger.debug("ManageClient.create_key LEAVE")
        return res

    def delete_key(
        self,
        project_id: str,
        key_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Deletes a key.

        Reference:
        https://developers.deepgram.com/reference/delete-key
        """
        self._logger.debug("ManageClient.delete_key ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/keys/{key_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("key_id: %s", key_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.delete(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("delete_key succeeded")
        self._logger.debug("ManageClient.delete_key LEAVE")
        return res

    # members
    def list_members(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> MembersResponse:
        """
        Please see get_members for more information.
        """
        return self.get_members(
            project_id, timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    def get_members(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> MembersResponse:
        """
        Gets a list of members for a project.

        Reference:
        https://developers.deepgram.com/reference/get-members
        """
        self._logger.debug("ManageClient.get_members ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/members"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = MembersResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_members succeeded")
        self._logger.debug("ManageClient.get_members LEAVE")
        return res

    def remove_member(
        self,
        project_id: str,
        member_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Removes a member from a project.

        Reference:
        https://developers.deepgram.com/reference/remove-member
        """
        self._logger.debug("ManageClient.remove_member ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/members/{member_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("member_id: %s", member_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.delete(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("remove_member succeeded")
        self._logger.debug("ManageClient.remove_member LEAVE")
        return res

    # scopes
    def get_member_scopes(
        self,
        project_id: str,
        member_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> ScopesResponse:
        """
        Gets a list of scopes for a member.

        Reference:
        https://developers.deepgram.com/reference/get-member-scopes
        """
        self._logger.debug("ManageClient.get_member_scopes ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/members/{member_id}/scopes"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("member_id: %s", member_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = ScopesResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_member_scopes succeeded")
        self._logger.debug("ManageClient.get_member_scopes LEAVE")
        return res

    def update_member_scope(
        self,
        project_id: str,
        member_id: str,
        options: Union[Dict, ScopeOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Updates a member's scopes.

        Reference:
        https://developers.deepgram.com/reference/update-scope
        """
        self._logger.debug("ManageClient.update_member_scope ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/members/{member_id}/scopes"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, ScopeOptions):
            self._logger.info("ScopeOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.put(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("update_member_scope succeeded")
        self._logger.debug("ManageClient.update_member_scope LEAVE")
        return res

    # invites
    def list_invites(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> InvitesResponse:
        """
        Please see get_invites for more information.
        """
        return self.get_invites(
            project_id, timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    def get_invites(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> InvitesResponse:
        """
        Gets a list of invites for a project.

        Reference:
        https://developers.deepgram.com/reference/list-invites
        """
        self._logger.debug("ManageClient.get_invites ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/invites"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = InvitesResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_invites succeeded")
        self._logger.debug("ManageClient.get_invites LEAVE")
        return res

    def send_invite_options(
        self,
        project_id: str,
        options: Union[Dict, InviteOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Sends an invite to a project.

        Reference:
        https://developers.deepgram.com/reference/send-invite
        """
        self._logger.debug("ManageClient.send_invite_options ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/invites"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, InviteOptions):
            self._logger.info("InviteOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.post(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("send_invite_options succeeded")
        self._logger.debug("ManageClient.send_invite_options LEAVE")
        return res

    def send_invite(
        self,
        project_id: str,
        email: str,
        scope="member",
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Sends an invite to a project.

        Reference:
        https://developers.deepgram.com/reference/send-invite
        """
        self._logger.debug("ManageClient.send_invite ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/invites"
        options = {
            "email": email,
            "scope": scope,
        }
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.post(
            url, json=options, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("send_invite succeeded")
        self._logger.debug("ManageClient.send_invite LEAVE")
        return res

    def delete_invite(
        self,
        project_id: str,
        email: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Deletes an invite from a project.

        Reference:
        https://developers.deepgram.com/reference/delete-invite
        """
        self._logger.debug("ManageClient.delete_invite ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/invites/{email}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("email: %s", email)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.delete(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("delete_invite succeeded")
        self._logger.debug("ManageClient.delete_invite LEAVE")
        return res

    def leave_project(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Message:
        """
        Leaves a project.

        Reference:
        https://developers.deepgram.com/reference/leave-project
        """
        self._logger.debug("ManageClient.leave_project ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/leave"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.delete(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Message.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("leave_project succeeded")
        self._logger.debug("ManageClient.leave_project LEAVE")
        return res

    # usage
    def get_usage_requests(
        self,
        project_id: str,
        options: Union[Dict, UsageRequestOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> UsageRequestsResponse:
        """
        Gets a list of usage requests for a project.

        Reference:
        https://developers.deepgram.com/reference/get-all-requests
        """
        self._logger.debug("ManageClient.get_usage_requests ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/requests"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, UsageRequestOptions):
            self._logger.info("UsageRequestOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url,
            options=options,
            timeout=timeout,
            addons=addons,
            headers=headers,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = UsageRequestsResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_usage_requests succeeded")
        self._logger.debug("ManageClient.get_usage_requests LEAVE")
        return res

    def get_usage_request(
        self,
        project_id: str,
        request_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> UsageRequest:
        """
        Gets details for a specific usage request.

        Reference:
        https://developers.deepgram.com/reference/get-request
        """
        self._logger.debug("ManageClient.get_usage_request ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/requests/{request_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("request_id: %s", request_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)

        # convert str to JSON to check response field
        json_result = json.loads(result)
        if json_result.get("response") is None:
            raise DeepgramError(
                "Response is not available yet. Please try again later."
            )

        res = UsageRequest.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_usage_request succeeded")
        self._logger.debug("ManageClient.get_usage_request LEAVE")
        return res

    def get_usage_summary(
        self,
        project_id: str,
        options: Union[Dict, UsageSummaryOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> UsageSummaryResponse:
        """
        Gets a summary of usage for a project.

        Reference:
        https://developers.deepgram.com/reference/summarize-usage
        """
        self._logger.debug("ManageClient.get_usage_summary ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/usage"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, UsageSummaryOptions):
            self._logger.info("UsageSummaryOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url,
            options=options,
            timeout=timeout,
            addons=addons,
            headers=headers,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = UsageSummaryResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_usage_summary succeeded")
        self._logger.debug("ManageClient.get_usage_summary LEAVE")
        return res

    def get_usage_fields(
        self,
        project_id: str,
        options: Union[Dict, UsageFieldsOptions],
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> UsageFieldsResponse:
        """
        Gets a list of usage fields for a project.

        Reference:
        https://developers.deepgram.com/reference/get-fields
        """
        self._logger.debug("ManageClient.get_usage_fields ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/usage/fields"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        if isinstance(options, UsageFieldsOptions):
            self._logger.info("UsageFieldsOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url,
            options=options,
            timeout=timeout,
            addons=addons,
            headers=headers,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = UsageFieldsResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_usage_fields succeeded")
        self._logger.debug("ManageClient.get_usage_fields LEAVE")
        return res

    # balances
    def list_balances(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> BalancesResponse:
        """
        Please see get_balances for more information.
        """
        return self.get_balances(
            project_id, timeout=timeout, addons=addons, headers=headers, **kwargs
        )

    def get_balances(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> BalancesResponse:
        """
        Gets a list of balances for a project.

        Reference:
        https://developers.deepgram.com/reference/get-all-balances
        """
        self._logger.debug("ManageClient.get_balances ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/balances"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = BalancesResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_balances succeeded")
        self._logger.debug("ManageClient.get_balances LEAVE")
        return res

    def get_balance(
        self,
        project_id: str,
        balance_id: str,
        timeout: Optional[httpx.Timeout] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs,
    ) -> Balance:
        """
        Gets details for a specific balance.

        Reference:
        https://developers.deepgram.com/reference/get-balance
        """
        self._logger.debug("ManageClient.get_balance ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/balances/{balance_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("balance_id: %s", balance_id)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = self.get(
            url, timeout=timeout, addons=addons, headers=headers, **kwargs
        )
        self._logger.info("json: %s", result)
        res = Balance.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_balance succeeded")
        self._logger.debug("ManageClient.get_balance LEAVE")
        return res

    # pylint: enable=too-many-positional-arguments



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import AnalyzeClient, AsyncAnalyzeClient
from .client import ReadClient, AsyncReadClient
from .client import AnalyzeOptions
from .client import (
    # common
    UrlSource,
    TextSource,
    BufferSource,
    StreamSource,
    FileSource,
    # unique
    AnalyzeStreamSource,
    AnalyzeSource,
)
from .client import (
    AsyncAnalyzeResponse,
    SyncAnalyzeResponse,
    AnalyzeResponse,
    # shared
    Average,
    Intent,
    Intents,
    IntentsInfo,
    Segment,
    SentimentInfo,
    Sentiment,
    Sentiments,
    SummaryInfo,
    Topic,
    Topics,
    TopicsInfo,
    # unique
    AnalyzeMetadata,
    AnalyzeResults,
    AnalyzeSummary,
)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import SpeakWebSocketClient, SpeakWSClient
from .async_client import AsyncSpeakWebSocketClient, AsyncSpeakWSClient
from .response import (
    #### top level
    MetadataResponse,
    FlushedResponse,
    ClearedResponse,
    WarningResponse,
    #### shared
    OpenResponse,
    CloseResponse,
    UnhandledResponse,
    ErrorResponse,
)
from .options import SpeakWSOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import logging
from typing import Dict, Union, Optional

import httpx

from .....utils import verboselogs
from .....options import DeepgramClientOptions
from ....common import AbstractAsyncRestClient
from ....common import DeepgramError, DeepgramTypeError

from .helpers import is_buffer_source, is_readstream_source, is_url_source
from .options import (
    ListenRESTOptions,
    PrerecordedOptions,
    FileSource,
    UrlSource,
)
from .response import AsyncPrerecordedResponse, PrerecordedResponse


class AsyncListenRESTClient(AbstractAsyncRestClient):
    """
    A client class for handling pre-recorded audio data.
    Provides methods for transcribing audio from URLs and files.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions

    def __init__(self, config: DeepgramClientOptions):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)
        self._config = config
        super().__init__(config)

    # pylint: disable=too-many-positional-arguments

    async def transcribe_url(
        self,
        source: UrlSource,
        options: Optional[Union[Dict, ListenRESTOptions]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
        **kwargs,
    ) -> Union[AsyncPrerecordedResponse, PrerecordedResponse]:
        """
        Transcribes audio from a URL source.

        Args:
            source (UrlSource): The URL source of the audio to transcribe.
            options (ListenRESTOptions): Additional options for the transcription (default is None).
            endpoint (str): The API endpoint for the transcription (default is "v1/listen").

        Returns:
            PrerecordedResponse: An object containing the transcription result.

        Raises:
            DeepgramTypeError: Raised for known API errors.
        """
        self._logger.debug("ListenRESTClient.transcribe_url ENTER")

        if (
            isinstance(options, dict)
            and "callback" in options
            and options["callback"] is not None
        ) or (isinstance(options, ListenRESTOptions) and options.callback is not None):
            self._logger.debug("ListenRESTClient.transcribe_url LEAVE")
            return await self.transcribe_url_callback(
                source,
                callback=options["callback"],
                options=options,
                addons=addons,
                headers=headers,
                timeout=timeout,
                endpoint=endpoint,
                **kwargs,
            )

        url = f"{self._config.url}/{endpoint}"
        if is_url_source(source):
            body = source
        else:
            self._logger.error("Unknown transcription source type")
            self._logger.debug("ListenRESTClient.transcribe_url LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, ListenRESTOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("ListenRESTClient.transcribe_url LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self._logger.info("url: %s", url)
        self._logger.info("source: %s", source)
        if isinstance(options, ListenRESTOptions):
            self._logger.info("ListenRESTOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = await self.post(
            url,
            options=options,
            addons=addons,
            headers=headers,
            json=body,
            timeout=timeout,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = PrerecordedResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("transcribe_url succeeded")
        self._logger.debug("ListenRESTClient.transcribe_url LEAVE")
        return res

    async def transcribe_url_callback(
        self,
        source: UrlSource,
        callback: str,
        options: Optional[Union[Dict, ListenRESTOptions]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
        **kwargs,
    ) -> AsyncPrerecordedResponse:
        """
        Transcribes audio from a URL source and sends the result to a callback URL.

        Args:
            source (UrlSource): The URL source of the audio to transcribe.
            callback (str): The callback URL where the transcription results will be sent.
            options (ListenRESTOptions): Additional options for the transcription (default is None).
            endpoint (str): The API endpoint for the transcription (default is "v1/listen").

        Returns:
            AsyncPrerecordedResponse: An object containing the request_id or an error message.

        Raises:
            DeepgramTypeError: Raised for known API errors.
        """
        self._logger.debug("ListenRESTClient.transcribe_url_callback ENTER")

        url = f"{self._config.url}/{endpoint}"
        if options is None:
            options = {}
        if isinstance(options, ListenRESTOptions):
            options.callback = callback
        else:
            options["callback"] = callback
        if is_url_source(source):
            body = source
        else:
            self._logger.error("Unknown transcription source type")
            self._logger.debug("ListenRESTClient.transcribe_url_callback LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, ListenRESTOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("ListenRESTClient.transcribe_url_callback LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self._logger.info("url: %s", url)
        self._logger.info("source: %s", source)
        if isinstance(options, ListenRESTOptions):
            self._logger.info("ListenRESTOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = await self.post(
            url,
            options=options,
            addons=addons,
            headers=headers,
            json=body,
            timeout=timeout,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = AsyncPrerecordedResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("transcribe_url_callback succeeded")
        self._logger.debug("ListenRESTClient.transcribe_url_callback LEAVE")
        return res

    async def transcribe_file(
        self,
        source: FileSource,
        options: Optional[Union[Dict, ListenRESTOptions]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
        **kwargs,
    ) -> Union[AsyncPrerecordedResponse, PrerecordedResponse]:
        """
        Transcribes audio from a local file source.

        Args:
            source (FileSource): The local file source of the audio to transcribe.
            options (ListenRESTOptions): Additional options for the transcription (default is None).
            endpoint (str): The API endpoint for the transcription (default is "v1/listen").

        Returns:
            PrerecordedResponse: An object containing the transcription result or an error message.

        Raises:
            DeepgramTypeError: Raised for known API errors.
        """
        self._logger.debug("ListenRESTClient.transcribe_file ENTER")

        if (
            isinstance(options, dict)
            and "callback" in options
            and options["callback"] is not None
        ) or (isinstance(options, ListenRESTOptions) and options.callback is not None):
            self._logger.debug("ListenRESTClient.transcribe_file LEAVE")
            return await self.transcribe_file_callback(
                source,
                callback=options["callback"],
                options=options,
                addons=addons,
                headers=headers,
                timeout=timeout,
                endpoint=endpoint,
                **kwargs,
            )

        url = f"{self._config.url}/{endpoint}"
        if is_buffer_source(source):
            body = source["buffer"]  # type: ignore
        elif is_readstream_source(source):
            body = source["stream"]  # type: ignore
        else:
            self._logger.error("Unknown transcription source type")
            self._logger.debug("ListenRESTClient.transcribe_file LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, ListenRESTOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("ListenRESTClient.transcribe_file LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self._logger.info("url: %s", url)
        if isinstance(options, ListenRESTOptions):
            self._logger.info("ListenRESTOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = await self.post(
            url,
            options=options,
            addons=addons,
            headers=headers,
            content=body,
            timeout=timeout,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = PrerecordedResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("transcribe_file succeeded")
        self._logger.debug("ListenRESTClient.transcribe_file LEAVE")
        return res

    async def transcribe_file_callback(
        self,
        source: FileSource,
        callback: str,
        options: Optional[Union[Dict, ListenRESTOptions]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
        **kwargs,
    ) -> AsyncPrerecordedResponse:
        """
        Transcribes audio from a local file source and sends the result to a callback URL.

        Args:
            source (FileSource): The local file source of the audio to transcribe.
            callback (str): The callback URL where the transcription results will be sent.
            options (ListenRESTOptions): Additional options for the transcription (default is None).
            endpoint (str): The API endpoint for the transcription (default is "v1/listen").

        Returns:
            AsyncPrerecordedResponse: An object containing the request_id or an error message.

        Raises:
            DeepgramTypeError: Raised for known API errors.
        """
        self._logger.debug("ListenRESTClient.transcribe_file_callback ENTER")

        url = f"{self._config.url}/{endpoint}"
        if options is None:
            options = {}
        if isinstance(options, ListenRESTOptions):
            options.callback = callback
        else:
            options["callback"] = callback
        if is_buffer_source(source):
            body = source["buffer"]  # type: ignore
        elif is_readstream_source(source):
            body = source["stream"]  # type: ignore
        else:
            self._logger.error("Unknown transcription source type")
            self._logger.debug("ListenRESTClient.transcribe_file_callback LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, ListenRESTOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("ListenRESTClient.transcribe_file_callback LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self._logger.info("url: %s", url)
        if isinstance(options, ListenRESTOptions):
            self._logger.info("ListenRESTOptions switching class -> dict")
            options = options.to_dict()
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        result = await self.post(
            url,
            options=options,
            addons=addons,
            headers=headers,
            content=body,
            timeout=timeout,
            **kwargs,
        )
        self._logger.info("json: %s", result)
        res = AsyncPrerecordedResponse.from_json(result)
        self._logger.verbose("result: %s", res)
        self._logger.notice("transcribe_file_callback succeeded")
        self._logger.debug("ListenRESTClient.transcribe_file_callback LEAVE")
        return res

    # pylint: enable=too-many-positional-arguments



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT
import json
import time
import logging
from typing import Dict, Union, Optional, cast, Any, Callable, Type
from datetime import datetime
import threading

from .....utils import verboselogs
from .....options import DeepgramClientOptions
from ...enums import LiveTranscriptionEvents
from ....common import AbstractSyncWebSocketClient
from ....common import DeepgramError

from .response import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
)
from .options import ListenWebSocketOptions

ONE_SECOND = 1
HALF_SECOND = 0.5
DEEPGRAM_INTERVAL = 5
PING_INTERVAL = 20


class ListenWebSocketClient(
    AbstractSyncWebSocketClient
):  # pylint: disable=too-many-instance-attributes
    """
    Client for interacting with Deepgram's live transcription services over WebSockets.

    This class provides methods to establish a WebSocket connection for live transcription and handle real-time transcription events.

    Args:
        config (DeepgramClientOptions): all the options for the client.
        thread_cls (Type[threading.Thread]): optional thread class to use for creating threads,
        defaults to threading.Thread. Useful for custom thread management like ContextVar support.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions
    _endpoint: str

    _lock_flush: threading.Lock
    _event_handlers: Dict[LiveTranscriptionEvents, list]

    _keep_alive_thread: Union[threading.Thread, None]
    _flush_thread: Union[threading.Thread, None]
    _last_datagram: Optional[datetime] = None

    _thread_cls: Type[threading.Thread]

    _kwargs: Optional[Dict] = None
    _addons: Optional[Dict] = None
    _options: Optional[Dict] = None
    _headers: Optional[Dict] = None

    def __init__(
        self,
        config: DeepgramClientOptions,
        thread_cls: Type[threading.Thread] = threading.Thread,
    ):
        if config is None:
            raise DeepgramError("Config is required")

        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)

        self._config = config
        self._endpoint = "v1/listen"

        self._flush_thread = None
        self._keep_alive_thread = None

        # auto flush
        self._last_datagram = None
        self._lock_flush = threading.Lock()

        self._thread_cls = thread_cls

        # init handlers
        self._event_handlers = {
            event: [] for event in LiveTranscriptionEvents.__members__.values()
        }

        # call the parent constructor
        super().__init__(
            config=self._config,
            endpoint=self._endpoint,
            thread_cls=self._thread_cls,
        )

    # pylint: disable=too-many-statements,too-many-branches
    def start(
        self,
        options: Optional[Union[ListenWebSocketOptions, Dict]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        members: Optional[Dict] = None,
        **kwargs,
    ) -> bool:
        """
        Starts the WebSocket connection for live transcription.
        """
        self._logger.debug("ListenWebSocketClient.start ENTER")
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        self._logger.info("members: %s", members)
        self._logger.info("kwargs: %s", kwargs)

        if isinstance(options, ListenWebSocketOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("ListenWebSocketClient.start LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self._addons = addons
        self._headers = headers

        # add "members" as members of the class
        if members is not None:
            self.__dict__.update(members)

        # set kwargs as members of the class
        if kwargs is not None:
            self._kwargs = kwargs
        else:
            self._kwargs = {}

        if isinstance(options, ListenWebSocketOptions):
            self._logger.info("ListenWebSocketOptions switching class -> dict")
            self._options = options.to_dict()
        elif options is not None:
            self._options = options
        else:
            self._options = {}

        try:
            # call parent start
            if (
                super().start(
                    self._options,
                    self._addons,
                    self._headers,
                    **dict(cast(Dict[Any, Any], self._kwargs)),
                )
                is False
            ):
                self._logger.error("ListenWebSocketClient.start failed")
                self._logger.debug("ListenWebSocketClient.start LEAVE")
                return False

            # debug the threads
            for thread in threading.enumerate():
                self._logger.debug("after running thread: %s", thread.name)
            self._logger.debug("number of active threads: %s", threading.active_count())

            # keepalive thread
            if self._config.is_keep_alive_enabled():
                self._logger.notice("keepalive is enabled")
                self._keep_alive_thread = self._thread_cls(target=self._keep_alive)
                self._keep_alive_thread.start()
            else:
                self._logger.notice("keepalive is disabled")

            # flush thread
            if self._config.is_auto_flush_reply_enabled():
                self._logger.notice("autoflush is enabled")
                self._flush_thread = self._thread_cls(target=self._flush)
                self._flush_thread.start()
            else:
                self._logger.notice("autoflush is disabled")

            # debug the threads
            for thread in threading.enumerate():
                self._logger.debug("after running thread: %s", thread.name)
            self._logger.debug("number of active threads: %s", threading.active_count())

            self._logger.notice("start succeeded")
            self._logger.debug("ListenWebSocketClient.start LEAVE")
            return True

        except Exception as e:  # pylint: disable=broad-except
            self._logger.error(
                "WebSocketException in ListenWebSocketClient.start: %s", e
            )
            self._logger.debug("ListenWebSocketClient.start LEAVE")
            if self._config.options.get("termination_exception_connect") is True:
                raise e
            return False

    # pylint: enable=too-many-statements,too-many-branches

    def on(
        self, event: LiveTranscriptionEvents, handler: Callable
    ) -> None:  # registers event handlers for specific events
        """
        Registers event handlers for specific events.
        """
        self._logger.info("event subscribed: %s", event)
        if event in LiveTranscriptionEvents.__members__.values() and callable(handler):
            self._event_handlers[event].append(handler)

    def _emit(self, event: LiveTranscriptionEvents, *args, **kwargs) -> None:
        """
        Emits events to the registered event handlers.
        """
        self._logger.debug("ListenWebSocketClient._emit ENTER")
        self._logger.debug("callback handlers for: %s", event)

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("after running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.debug("callback handlers for: %s", event)
        for handler in self._event_handlers[event]:
            handler(self, *args, **kwargs)

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("after running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.debug("ListenWebSocketClient._emit LEAVE")

    # pylint: disable=too-many-return-statements,too-many-statements,too-many-locals,too-many-branches
    def _process_text(self, message: str) -> None:
        """
        Processes messages received over the WebSocket connection.
        """
        self._logger.debug("ListenWebSocketClient._process_text ENTER")

        try:
            if len(message) == 0:
                self._logger.debug("message is empty")
                self._logger.debug("ListenWebSocketClient._process_text LEAVE")
                return

            data = json.loads(message)
            response_type = data.get("type")
            self._logger.debug("response_type: %s, data: %s", response_type, data)

            match response_type:
                case LiveTranscriptionEvents.Open:
                    open_result: OpenResponse = OpenResponse.from_json(message)
                    self._logger.verbose("OpenResponse: %s", open_result)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Open),
                        open=open_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.Transcript:
                    msg_result: LiveResultResponse = LiveResultResponse.from_json(
                        message
                    )
                    self._logger.verbose("LiveResultResponse: %s", msg_result)

                    #  auto flush
                    if self._config.is_inspecting_listen():
                        inspect_res = self._inspect(msg_result)
                        if not inspect_res:
                            self._logger.error("inspect_res failed")

                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Transcript),
                        result=msg_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.Metadata:
                    meta_result: MetadataResponse = MetadataResponse.from_json(message)
                    self._logger.verbose("MetadataResponse: %s", meta_result)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Metadata),
                        metadata=meta_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.SpeechStarted:
                    ss_result: SpeechStartedResponse = SpeechStartedResponse.from_json(
                        message
                    )
                    self._logger.verbose("SpeechStartedResponse: %s", ss_result)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.SpeechStarted),
                        speech_started=ss_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.UtteranceEnd:
                    ue_result: UtteranceEndResponse = UtteranceEndResponse.from_json(
                        message
                    )
                    self._logger.verbose("UtteranceEndResponse: %s", ue_result)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.UtteranceEnd),
                        utterance_end=ue_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.Close:
                    close_result: CloseResponse = CloseResponse.from_json(message)
                    self._logger.verbose("CloseResponse: %s", close_result)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Close),
                        close=close_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case LiveTranscriptionEvents.Error:
                    err_error: ErrorResponse = ErrorResponse.from_json(message)
                    self._logger.verbose("ErrorResponse: %s", err_error)
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Error),
                        error=err_error,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case _:
                    self._logger.warning(
                        "Unknown Message: response_type: %s, data: %s",
                        response_type,
                        data,
                    )
                    unhandled_error: UnhandledResponse = UnhandledResponse(
                        type=LiveTranscriptionEvents(LiveTranscriptionEvents.Unhandled),
                        raw=message,
                    )
                    self._emit(
                        LiveTranscriptionEvents(LiveTranscriptionEvents.Unhandled),
                        unhandled=unhandled_error,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )

            self._logger.notice("_process_text Succeeded")
            self._logger.debug("SpeakStreamClient._process_text LEAVE")

        except Exception as e:  # pylint: disable=broad-except
            self._logger.error(
                "Exception in ListenWebSocketClient._process_text: %s", e
            )
            e_error: ErrorResponse = ErrorResponse(
                "Exception in ListenWebSocketClient._process_text",
                f"{e}",
                "Exception",
            )
            self._logger.error(
                "Exception in ListenWebSocketClient._process_text: %s", str(e)
            )
            self._emit(
                LiveTranscriptionEvents(LiveTranscriptionEvents.Error),
                e_error,
                **dict(cast(Dict[Any, Any], self._kwargs)),
            )

            # signal exit and close
            super()._signal_exit()

            self._logger.debug("ListenWebSocketClient._process_text LEAVE")

            if self._config.options.get("termination_exception") is True:
                raise
            return

    # pylint: enable=too-many-return-statements,too-many-statements

    def _process_binary(self, message: bytes) -> None:
        raise NotImplementedError("no _process_binary method should be called")

    # pylint: disable=too-many-return-statements
    def _keep_alive(self) -> None:
        self._logger.debug("ListenWebSocketClient._keep_alive ENTER")

        counter = 0
        while True:
            try:
                counter += 1
                self._exit_event.wait(timeout=ONE_SECOND)

                if self._exit_event.is_set():
                    self._logger.notice("_keep_alive exiting gracefully")
                    self._logger.debug("ListenWebSocketClient._keep_alive LEAVE")
                    return

                # deepgram keepalive
                if counter % DEEPGRAM_INTERVAL == 0:
                    self.keep_alive()

            except Exception as e:  # pylint: disable=broad-except
                self._logger.error(
                    "Exception in ListenWebSocketClient._keep_alive: %s", e
                )
                e_error: ErrorResponse = ErrorResponse(
                    "Exception in ListenWebSocketClient._keep_alive",
                    f"{e}",
                    "Exception",
                )
                self._logger.error(
                    "Exception in ListenWebSocketClient._keep_alive: %s", str(e)
                )
                self._emit(
                    LiveTranscriptionEvents(LiveTranscriptionEvents.Error),
                    e_error,
                    **dict(cast(Dict[Any, Any], self._kwargs)),
                )

                # signal exit and close
                super()._signal_exit()

                self._logger.debug("ListenWebSocketClient._keep_alive LEAVE")

                if self._config.options.get("termination_exception") is True:
                    raise
                return

    ## pylint: disable=too-many-return-statements,too-many-statements
    def _flush(self) -> None:
        self._logger.debug("ListenWebSocketClient._flush ENTER")

        delta_in_ms_str = self._config.options.get("auto_flush_reply_delta")
        if delta_in_ms_str is None:
            self._logger.error("auto_flush_reply_delta is None")
            self._logger.debug("ListenWebSocketClient._flush LEAVE")
            return
        delta_in_ms = float(delta_in_ms_str)

        _flush_event = threading.Event()
        while True:
            try:
                _flush_event.wait(timeout=HALF_SECOND)

                if self._exit_event.is_set():
                    self._logger.notice("_flush exiting gracefully")
                    self._logger.debug("ListenWebSocketClient._flush LEAVE")
                    return

                with self._lock_flush:
                    if self._last_datagram is None:
                        self._logger.debug("AutoFlush last_datagram is None")
                        continue

                    delta = datetime.now() - self._last_datagram
                    diff_in_ms = delta.total_seconds() * 1000
                    self._logger.debug("AutoFlush delta: %f", diff_in_ms)
                    if diff_in_ms < delta_in_ms:
                        self._logger.debug("AutoFlush delta is less than threshold")
                        continue

                    with self._lock_flush:
                        self._last_datagram = None
                    self.finalize()

            except Exception as e:  # pylint: disable=broad-except
                self._logger.error("Exception in ListenWebSocketClient._flush: %s", e)
                e_error: ErrorResponse = ErrorResponse(
                    "Exception in ListenWebSocketClient._flush",
                    f"{e}",
                    "Exception",
                )
                self._logger.error(
                    "Exception in ListenWebSocketClient._flush: %s", str(e)
                )
                self._emit(
                    LiveTranscriptionEvents(LiveTranscriptionEvents.Error),
                    e_error,
                    **dict(cast(Dict[Any, Any], self._kwargs)),
                )

                # signal exit and close
                super()._signal_exit()

                self._logger.debug("ListenWebSocketClient._flush LEAVE")

                if self._config.options.get("termination_exception") is True:
                    raise
                return

    # pylint: enable=too-many-return-statements

    def keep_alive(self) -> bool:
        """
        Sends a KeepAlive message
        """
        self._logger.spam("ListenWebSocketClient.keep_alive ENTER")

        self._logger.notice("Sending KeepAlive...")
        ret = self.send(json.dumps({"type": "KeepAlive"}))

        if not ret:
            self._logger.error("keep_alive failed")
            self._logger.spam("ListenWebSocketClient.keep_alive LEAVE")
            return False

        self._logger.notice("keep_alive succeeded")
        self._logger.spam("ListenWebSocketClient.keep_alive LEAVE")

        return True

    def finalize(self) -> bool:
        """
        Finalizes the Transcript connection by flushing it
        """
        self._logger.spam("ListenWebSocketClient.finalize ENTER")

        self._logger.notice("Sending Finalize...")
        ret = self.send(json.dumps({"type": "Finalize"}))

        if not ret:
            self._logger.error("finalize failed")
            self._logger.spam("ListenWebSocketClient.finalize LEAVE")
            return False

        self._logger.notice("finalize succeeded")
        self._logger.spam("ListenWebSocketClient.finalize LEAVE")

        return True

    def _close_message(self) -> bool:
        return self.send(json.dumps({"type": "CloseStream"}))

    # closes the WebSocket connection gracefully
    def finish(self) -> bool:
        """
        Closes the WebSocket connection gracefully.
        """
        self._logger.spam("ListenWebSocketClient.finish ENTER")

        # call parent finish
        if super().finish() is False:
            self._logger.error("ListenWebSocketClient.finish failed")

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("before running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        # stop the threads
        self._logger.verbose("cancelling tasks...")
        if self._flush_thread is not None:
            self._flush_thread.join()
            self._flush_thread = None
            self._logger.notice("processing _flush_thread thread joined")

        if self._keep_alive_thread is not None:
            self._keep_alive_thread.join()
            self._keep_alive_thread = None
            self._logger.notice("processing _keep_alive_thread thread joined")

        if self._listen_thread is not None:
            self._listen_thread.join()
            self._listen_thread = None
        self._logger.notice("listening thread joined")

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("before running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.notice("finish succeeded")
        self._logger.spam("ListenWebSocketClient.finish LEAVE")
        return True

    def _inspect(self, msg_result: LiveResultResponse) -> bool:
        # auto flush_inspect is generically used to track any messages you might want to snoop on
        # place additional logic here to inspect messages of interest

        # for auto flush functionality
        # set the last datagram
        sentence = msg_result.channel.alternatives[0].transcript
        if len(sentence) == 0:
            return True

        if msg_result.is_final:
            with self._lock_flush:
                self._logger.debug("AutoFlush is_final received")
                self._last_datagram = None
        else:
            with self._lock_flush:
                self._last_datagram = datetime.now()
                self._logger.debug(
                    "AutoFlush interim received: %s",
                    str(self._last_datagram),
                )

        return True



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import logging
from typing import Optional

import httpx

from ....utils import verboselogs
from ....options import DeepgramClientOptions
from ...common import AbstractSyncRestClient


class SelfHostedClient(AbstractSyncRestClient):
    """
    Client for interacting with Deepgram's on-premises API.

    This class provides methods to manage and interact with on-premises projects and distribution credentials.

    Args:
        config (DeepgramClientOptions): all the options for the client.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions
    _endpoint: str

    def __init__(self, config: DeepgramClientOptions):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)
        self._config = config
        self._endpoint = "v1/projects"
        super().__init__(config)

    def list_onprem_credentials(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        List all on-premises distribution credentials for a project.
        """
        return self.list_selfhosted_credentials(project_id, timeout=timeout, **kwargs)

    def list_selfhosted_credentials(
        self,
        project_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        List all on-premises distribution credentials for a project.
        """
        self._logger.debug("SelfHostedClient.list_selfhosted_credentials ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/selfhosted/distribution/credentials"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        res = self.get(url, timeout=timeout, **kwargs)
        self._logger.verbose("result: %s", res)
        self._logger.notice("list_selfhosted_credentials succeeded")
        self._logger.debug("SelfHostedClient.list_selfhosted_credentials LEAVE")
        return res

    def get_onprem_credentials(
        self,
        project_id: str,
        distribution_credentials_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Get a specific on-premises distribution credential for a project.
        """
        return self.get_selfhosted_credentials(
            project_id, distribution_credentials_id, timeout=timeout, **kwargs
        )

    def get_selfhosted_credentials(
        self,
        project_id: str,
        distribution_credentials_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Get a specific on-premises distribution credential for a project.
        """
        self._logger.debug("SelfHostedClient.get_selfhosted_credentials ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/selfhosted/distribution/credentials/{distribution_credentials_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info(
            "distribution_credentials_id: %s", distribution_credentials_id
        )
        res = self.get(url, timeout=timeout, **kwargs)
        self._logger.verbose("result: %s", res)
        self._logger.notice("get_selfhosted_credentials succeeded")
        self._logger.debug("SelfHostedClient.get_selfhosted_credentials LEAVE")
        return res

    def create_onprem_credentials(
        self,
        project_id: str,
        options,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Create a new on-premises distribution credential for a project.
        """
        return self.create_selfhosted_credentials(
            project_id, options, timeout=timeout, **kwargs
        )

    def create_selfhosted_credentials(
        self,
        project_id: str,
        options,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Create a new on-premises distribution credential for a project.
        """
        self._logger.debug("SelfHostedClient.create_selfhosted_credentials ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/selfhosted/distribution/credentials/"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("options: %s", options)
        res = self.post(url, json=options, timeout=timeout, **kwargs)
        self._logger.verbose("result: %s", res)
        self._logger.notice("create_selfhosted_credentials succeeded")
        self._logger.debug("SelfHostedClient.create_selfhosted_credentials LEAVE")
        return res

    def delete_onprem_credentials(
        self,
        project_id: str,
        distribution_credentials_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Delete an on-premises distribution credential for a project.
        """
        return self.delete_selfhosted_credentials(
            project_id, distribution_credentials_id, timeout=timeout, **kwargs
        )

    def delete_selfhosted_credentials(
        self,
        project_id: str,
        distribution_credentials_id: str,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ):
        """
        Delete an on-premises distribution credential for a project.
        """
        self._logger.debug("SelfHostedClient.delete_selfhosted_credentials ENTER")
        url = f"{self._config.url}/{self._endpoint}/{project_id}/selfhosted/distribution/credentials/{distribution_credentials_id}"
        self._logger.info("url: %s", url)
        self._logger.info("project_id: %s", project_id)
        self._logger.info("distrbution_credentials_id: %s", distribution_credentials_id)
        res = self.delete(url, timeout=timeout, **kwargs)
        self._logger.verbose("result: %s", res)
        self._logger.notice("delete_selfhosted_credentials succeeded")
        self._logger.debug("SelfHostedClient.delete_selfhosted_credentials LEAVE")
        return res



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import SelfHostedClient as SelfHostedClientLatest
from .v1.async_client import AsyncSelfHostedClient as AsyncSelfHostedClientLatest


# The client.py points to the current supported version in the SDK.
# Older versions are supported in the SDK for backwards compatibility.

SelfHostedClient = SelfHostedClientLatest
AsyncSelfHostedClient = AsyncSelfHostedClientLatest
OnPremClient = SelfHostedClientLatest
AsyncOnPremClient = AsyncSelfHostedClientLatest



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .microphone import Microphone
from .microphone import DeepgramMicrophoneError
from .microphone import (
    LOGGING as INPUT_LOGGING,
    CHANNELS as INPUT_CHANNELS,
    RATE as INPUT_RATE,
    CHUNK as INPUT_CHUNK,
)

from .speaker import Speaker
from .speaker import DeepgramSpeakerError
from .speaker import (
    LOGGING as OUTPUT_LOGGING,
    CHANNELS as OUTPUT_CHANNELS,
    RATE as OUTPUT_RATE,
    CHUNK as OUTPUT_CHUNK,
    PLAYBACK_DELTA as OUTPUT_PLAYBACK_DELTA,
)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import SelfHostedClient
from .async_client import AsyncSelfHostedClient



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import Optional
from importlib import import_module
import os
import logging
import deprecation  # type: ignore

from . import __version__
from .utils import verboselogs

# common
# pylint: disable=unused-import
from .clients import (
    TextSource,
    BufferSource,
    StreamSource,
    FileSource,
    UrlSource,
)
from .clients import BaseResponse
from .clients import (
    Average,
    Intent,
    Intents,
    IntentsInfo,
    Segment,
    SentimentInfo,
    Sentiment,
    Sentiments,
    SummaryInfo,
    Topic,
    Topics,
    TopicsInfo,
)
from .clients import (
    ModelInfo,
    Hit,
    Search,
)
from .clients import (
    OpenResponse,
    CloseResponse,
    UnhandledResponse,
    ErrorResponse,
)
from .clients import (
    DeepgramError,
    DeepgramTypeError,
    DeepgramModuleError,
    DeepgramApiError,
    DeepgramUnknownApiError,
)

# listen client
from .clients import ListenRouter, ReadRouter, SpeakRouter, AgentRouter

# speech-to-text
from .clients import LiveClient, AsyncLiveClient  # backward compat
from .clients import (
    ListenWebSocketClient,
    AsyncListenWebSocketClient,
)
from .clients import (
    ListenWebSocketOptions,
    LiveOptions,
    LiveTranscriptionEvents,
)

# live client responses
from .clients import (
    #### top level
    LiveResultResponse,
    ListenWSMetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    #### common websocket response
    # OpenResponse,
    # CloseResponse,
    # ErrorResponse,
    # UnhandledResponse,
    #### unique
    ListenWSMetadata,
    ListenWSAlternative,
    ListenWSChannel,
    ListenWSWord,
)

# prerecorded
from .clients import (
    # common
    # UrlSource,
    # BufferSource,
    # StreamSource,
    # TextSource,
    # FileSource,
    # unique
    PreRecordedStreamSource,
    PrerecordedSource,
    ListenRestSource,
)

from .clients import (
    PreRecordedClient,
    AsyncPreRecordedClient,
)  # backward compat
from .clients import (
    ListenRESTClient,
    AsyncListenRESTClient,
)
from .clients import (
    ListenRESTOptions,
    PrerecordedOptions,
)

# rest client responses
from .clients import (
    #### top level
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
    #### shared
    # Average,
    # Intent,
    # Intents,
    # IntentsInfo,
    # Segment,
    # SentimentInfo,
    # Sentiment,
    # Sentiments,
    # SummaryInfo,
    # Topic,
    # Topics,
    # TopicsInfo,
    #### between rest and websocket
    # ModelInfo,
    # Alternative,
    # Hit,
    # Search,
    # Channel,
    # Word,
    # unique
    Entity,
    ListenRESTMetadata,
    Paragraph,
    Paragraphs,
    ListenRESTResults,
    Sentence,
    Summaries,
    SummaryV1,
    SummaryV2,
    Translation,
    Utterance,
    Warning,
    ListenRESTAlternative,
    ListenRESTChannel,
    ListenRESTWord,
)

# read
from .clients import ReadClient, AsyncReadClient
from .clients import AnalyzeClient, AsyncAnalyzeClient
from .clients import (
    AnalyzeOptions,
    AnalyzeStreamSource,
    AnalyzeSource,
)

# read client responses
from .clients import (
    #### top level
    AsyncAnalyzeResponse,
    SyncAnalyzeResponse,
    AnalyzeResponse,
    #### shared
    # Average,
    # Intent,
    # Intents,
    # IntentsInfo,
    # Segment,
    # SentimentInfo,
    # Sentiment,
    # Sentiments,
    # SummaryInfo,
    # Topic,
    # Topics,
    # TopicsInfo,
    #### unique
    AnalyzeMetadata,
    AnalyzeResults,
    AnalyzeSummary,
)

# speak
## speak REST
from .clients import (
    #### top level
    SpeakRESTOptions,
    SpeakOptions,  # backward compat
    #### common
    # TextSource,
    # BufferSource,
    # StreamSource,
    # FileSource,
    #### unique
    SpeakSource,
    SpeakRestSource,
    SpeakRESTSource,
)

from .clients import (
    SpeakClient,  # backward compat
    SpeakRESTClient,
    AsyncSpeakRESTClient,
)

from .clients import (
    SpeakResponse,  # backward compat
    SpeakRESTResponse,
)

## speak WebSocket
from .clients import SpeakWebSocketEvents, SpeakWebSocketMessage

from .clients import (
    SpeakWSOptions,
)

from .clients import (
    SpeakWebSocketClient,
    AsyncSpeakWebSocketClient,
    SpeakWSClient,
    AsyncSpeakWSClient,
)

from .clients import (
    #### top level
    SpeakWSMetadataResponse,
    FlushedResponse,
    ClearedResponse,
    WarningResponse,
    #### common websocket response
    # OpenResponse,
    # CloseResponse,
    # UnhandledResponse,
    # ErrorResponse,
)

# auth client classes
from .clients import AuthRESTClient, AsyncAuthRESTClient

# auth client responses
from .clients import (
    GrantTokenResponse,
)

# manage client classes/input
from .clients import ManageClient, AsyncManageClient
from .clients import (
    ProjectOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)

# manage client responses
from .clients import (
    #### top level
    Message,
    ProjectsResponse,
    ModelResponse,
    ModelsResponse,
    MembersResponse,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageResponse,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    BalancesResponse,
    #### shared
    Project,
    STTDetails,
    TTSMetadata,
    TTSDetails,
    Member,
    Key,
    Invite,
    Config,
    STTUsageDetails,
    Callback,
    TokenDetail,
    SpeechSegment,
    TTSUsageDetails,
    STTTokens,
    TTSTokens,
    UsageSummaryResults,
    Resolution,
    UsageModel,
    Balance,
)

# on-prem
from .clients import (
    OnPremClient,
    AsyncOnPremClient,
    SelfHostedClient,
    AsyncSelfHostedClient,
)


# agent
from .clients import AgentWebSocketEvents

# websocket
from .clients import (
    AgentWebSocketClient,
    AsyncAgentWebSocketClient,
)

from .clients import (
    #### common websocket response
    # OpenResponse,
    # CloseResponse,
    # ErrorResponse,
    # UnhandledResponse,
    #### unique
    WelcomeResponse,
    SettingsAppliedResponse,
    ConversationTextResponse,
    UserStartedSpeakingResponse,
    AgentThinkingResponse,
    FunctionCallRequest,
    AgentStartedSpeakingResponse,
    AgentAudioDoneResponse,
    InjectionRefusedResponse,
)

from .clients import (
    # top level
    SettingsOptions,
    UpdatePromptOptions,
    UpdateSpeakOptions,
    InjectAgentMessageOptions,
    FunctionCallResponse,
    AgentKeepAlive,
    # sub level
    Listen,
    ListenProvider,
    Speak,
    SpeakProvider,
    Header,
    Item,
    Properties,
    Parameters,
    Function,
    Think,
    ThinkProvider,
    Agent,
    Input,
    Output,
    Audio,
    Endpoint,
)


# client errors and options
from .options import DeepgramClientOptions, ClientOptionsFromEnv
from .errors import DeepgramApiKeyError

# pylint: enable=unused-import


class Deepgram:  # pylint: disable=broad-exception-raised
    """
    The Deepgram class is no longer a class in version 3 of this SDK.
    """

    def __init__(self, *anything):
        raise Exception(
            """
            FATAL ERROR:
            You are attempting to instantiate a Deepgram object, which is no longer a class in version 3 of this SDK.

            To fix this issue:
                1. You need to revert to the previous version 2 of the SDK: pip install deepgram-sdk==2.12.0
                2. or, update your application's code to use version 3 of this SDK. See the README for more information.

            Things to consider:

                - This Version 3 of the SDK requires Python 3.10 or higher.
                  Older versions (3.9 and lower) of Python are nearing end-of-life: https://devguide.python.org/versions/
                  Understand the risks of using a version of Python nearing EOL.

                - Version 2 of the SDK will receive maintenance updates in the form of security fixes only.
                  No new features will be added to version 2 of the SDK.
            """
        )


class DeepgramClient:
    """
    Represents a client for interacting with the Deepgram API.

    This class provides a client for making requests to the Deepgram API with various configuration options.

    Attributes:
        api_key (str): The Deepgram API key used for authentication.
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Raises:
        DeepgramApiKeyError: If the API key is missing or invalid.

    Methods:
        listen: Returns a ListenClient instance for interacting with Deepgram's transcription services.

        manage: (Preferred) Returns a Threaded ManageClient instance for managing Deepgram resources.
        selfhosted: (Preferred) Returns an Threaded SelfHostedClient instance for interacting with Deepgram's on-premises API.

        asyncmanage: Returns an (Async) ManageClient instance for managing Deepgram resources.
        asyncselfhosted: Returns an (Async) SelfHostedClient instance for interacting with Deepgram's on-premises API.
    """

    _config: DeepgramClientOptions
    _logger: verboselogs.VerboseLogger

    def __init__(
        self,
        api_key: str = "",
        config: Optional[DeepgramClientOptions] = None,
    ):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())

        if api_key == "" and config is not None:
            self._logger.info("Attempting to set API key from config object")
            api_key = config.api_key
        if api_key == "":
            self._logger.info("Attempting to set API key from environment variable")
            api_key = os.getenv("DEEPGRAM_API_KEY", "")
        if api_key == "":
            self._logger.warning("WARNING: API key is missing")

        self.api_key = api_key
        if config is None:  # Use default configuration
            self._config = DeepgramClientOptions(self.api_key)
        else:
            config.set_apikey(self.api_key)
            self._config = config

    @property
    def listen(self):
        """
        Returns a Listen dot-notation router for interacting with Deepgram's transcription services.
        """
        return ListenRouter(self._config)

    @property
    def read(self):
        """
        Returns a Read dot-notation router for interacting with Deepgram's read services.
        """
        return ReadRouter(self._config)

    @property
    def speak(self):
        """
        Returns a Speak dot-notation router for interacting with Deepgram's speak services.
        """
        return SpeakRouter(self._config)

    @property
    @deprecation.deprecated(
        deprecated_in="3.4.0",
        removed_in="4.0.0",
        current_version=__version__,
        details="deepgram.asyncspeak is deprecated. Use deepgram.speak.asyncrest instead.",
    )
    def asyncspeak(self):
        """
        DEPRECATED: deepgram.asyncspeak is deprecated. Use deepgram.speak.asyncrest instead.
        """
        return self.Version(self._config, "asyncspeak")

    @property
    def manage(self):
        """
        Returns a ManageClient instance for managing Deepgram resources.
        """
        return self.Version(self._config, "manage")

    @property
    def asyncmanage(self):
        """
        Returns an AsyncManageClient instance for managing Deepgram resources.
        """
        return self.Version(self._config, "asyncmanage")

    @property
    def auth(self):
        """
        Returns an AuthRESTClient instance for managing short-lived tokens.
        """
        return self.Version(self._config, "auth")

    @property
    def asyncauth(self):
        """
        Returns an AsyncAuthRESTClient instance for managing short-lived tokens.
        """
        return self.Version(self._config, "asyncauth")

    @property
    @deprecation.deprecated(
        deprecated_in="3.4.0",
        removed_in="4.0.0",
        current_version=__version__,
        details="deepgram.onprem is deprecated. Use deepgram.speak.selfhosted instead.",
    )
    def onprem(self):
        """
        DEPRECATED: deepgram.onprem is deprecated. Use deepgram.speak.selfhosted instead.
        """
        return self.Version(self._config, "selfhosted")

    @property
    def selfhosted(self):
        """
        Returns an SelfHostedClient instance for interacting with Deepgram's on-premises API.
        """
        return self.Version(self._config, "selfhosted")

    @property
    @deprecation.deprecated(
        deprecated_in="3.4.0",
        removed_in="4.0.0",
        current_version=__version__,
        details="deepgram.asynconprem is deprecated. Use deepgram.speak.asyncselfhosted instead.",
    )
    def asynconprem(self):
        """
        DEPRECATED: deepgram.asynconprem is deprecated. Use deepgram.speak.asyncselfhosted instead.
        """
        return self.Version(self._config, "asyncselfhosted")

    @property
    def asyncselfhosted(self):
        """
        Returns an AsyncSelfHostedClient instance for interacting with Deepgram's on-premises API.
        """
        return self.Version(self._config, "asyncselfhosted")

    @property
    def agent(self):
        """
        Returns a Agent dot-notation router for interacting with Deepgram's speak services.
        """
        return AgentRouter(self._config)

    # INTERNAL CLASSES
    class Version:
        """
        Represents a version of the Deepgram API.
        """

        _logger: verboselogs.VerboseLogger
        _config: DeepgramClientOptions
        _parent: str

        def __init__(self, config, parent: str):
            self._logger = verboselogs.VerboseLogger(__name__)
            self._logger.addHandler(logging.StreamHandler())
            self._logger.setLevel(config.verbose)

            self._config = config
            self._parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self._parent:
        #         case "manage":
        #             return ManageClient(self._config)
        #         case "selfhosted":
        #             return SelfHostedClient(self._config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(self, version: str = ""):
            # pylint: disable-msg=too-many-statements
            """
            Returns a client for the specified version of the API.
            """
            self._logger.debug("Version.v ENTER")
            self._logger.info("version: %s", version)
            if len(version) == 0:
                self._logger.error("version is empty")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            parent = ""
            filename = ""
            classname = ""
            match self._parent:
                case "manage":
                    parent = "manage"
                    filename = "client"
                    classname = "ManageClient"
                case "asyncmanage":
                    parent = "manage"
                    filename = "async_client"
                    classname = "AsyncManageClient"
                case "asyncspeak":
                    return AsyncSpeakRESTClient(self._config)
                case "selfhosted":
                    parent = "selfhosted"
                    filename = "client"
                    classname = "SelfHostedClient"
                case "asyncselfhosted":
                    parent = "selfhosted"
                    filename = "async_client"
                    classname = "AsyncSelfHostedClient"
                case "auth":
                    parent = "auth"
                    filename = "client"
                    classname = "AuthRESTClient"
                case "asyncauth":
                    parent = "auth"
                    filename = "async_client"
                    classname = "AsyncAuthRESTClient"
                case _:
                    self._logger.error("parent unknown: %s", self._parent)
                    self._logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.{parent}.v{version}.{filename}"
            self._logger.info("path: %s", path)
            self._logger.info("classname: %s", classname)

            # import class
            mod = import_module(path)
            if mod is None:
                self._logger.error("module path is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, classname)
            if my_class is None:
                self._logger.error("my_class is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            my_class_instance = my_class(self._config)
            self._logger.notice("Version.v succeeded")
            self._logger.debug("Version.v LEAVE")
            return my_class_instance

        # pylint: enable-msg=too-many-statements



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from ...utils import verboselogs

# Constants for speaker
LOGGING = verboselogs.WARNING
TIMEOUT = 0.050
CHANNELS = 1
RATE = 16000
CHUNK = 8194

# Constants for speaker
PLAYBACK_DELTA = 2000



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import sys
import re
import os
from typing import Dict, Optional
import logging
import numbers

from deepgram import __version__
from .utils import verboselogs
from .errors import DeepgramApiKeyError


class DeepgramClientOptions:  # pylint: disable=too-many-instance-attributes
    """
    Represents options for configuring a Deepgram client.

    This class allows you to customize various options for interacting with the Deepgram API.

    Attributes:
        api_key: (Optional) A Deepgram API key used for authentication. Default uses the `DEEPGRAM_API_KEY` environment variable.
        url: (Optional) The URL used to interact with production, On-prem, and other Deepgram environments. Defaults to `api.deepgram.com`.
        verbose: (Optional) The logging level for the client. Defaults to `verboselogs.WARNING`.
        headers: (Optional) Headers for initializing the client.
        options: (Optional) Additional options for initializing the client.
    """

    _logger: verboselogs.VerboseLogger
    _inspect_listen: bool = False
    _inspect_speak: bool = False

    def __init__(
        self,
        api_key: str = "",
        url: str = "",
        verbose: int = verboselogs.WARNING,
        headers: Optional[Dict] = None,
        options: Optional[Dict] = None,
    ):  # pylint: disable=too-many-positional-arguments
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())

        if api_key is None:
            api_key = ""

        self.verbose = verbose
        self.api_key = api_key

        if headers is None:
            headers = {}
        self._update_headers(headers=headers)

        if len(url) == 0:
            url = "api.deepgram.com"
        self.url = self._get_url(url)

        if options is None:
            options = {}
        self.options = options

        if self.is_auto_flush_reply_enabled():
            self._inspect_listen = True
        if self.is_auto_flush_speak_enabled():
            self._inspect_speak = True

    def set_apikey(self, api_key: str):
        """
        set_apikey: Sets the API key for the client.

        Args:
            api_key: The Deepgram API key used for authentication.
        """
        self.api_key = api_key
        self._update_headers()

    def _get_url(self, url) -> str:
        if not re.match(r"^https?://", url, re.IGNORECASE):
            url = "https://" + url
        return url.strip("/")

    def _update_headers(self, headers: Optional[Dict] = None):
        self.headers = {}
        self.headers["Accept"] = "application/json"
        if self.api_key:
            self.headers["Authorization"] = f"Token {self.api_key}"
        elif "Authorization" in self.headers:
            del self.headers["Authorization"]
        self.headers[
            "User-Agent"
        ] = f"@deepgram/sdk/{__version__} python/{sys.version_info[1]}.{sys.version_info[2]}"
        # Overwrite / add any headers that were passed in
        if headers:
            self.headers.update(headers)

    def is_keep_alive_enabled(self) -> bool:
        """
        is_keep_alive_enabled: Returns True if the client is configured to keep the connection alive.
        """
        return self.options.get("keepalive", False) or self.options.get(
            "keep_alive", False
        )

    def is_auto_flush_reply_enabled(self) -> bool:
        """
        is_auto_flush_reply_enabled: Returns True if the client is configured to auto-flush for listen.
        """
        auto_flush_reply_delta = float(self.options.get("auto_flush_reply_delta", 0))
        return (
            isinstance(auto_flush_reply_delta, numbers.Number)
            and auto_flush_reply_delta > 0
        )

    def is_auto_flush_speak_enabled(self) -> bool:
        """
        is_auto_flush_speak_enabled: Returns True if the client is configured to auto-flush for speak.
        """
        auto_flush_speak_delta = float(self.options.get("auto_flush_speak_delta", 0))
        return (
            isinstance(auto_flush_speak_delta, numbers.Number)
            and auto_flush_speak_delta > 0
        )

    def is_inspecting_listen(self) -> bool:
        """
        is_inspecting_listen: Returns True if the client is inspecting listen messages.
        """
        return self._inspect_listen

    def is_inspecting_speak(self) -> bool:
        """
        is_inspecting_speak: Returns True if the client is inspecting speak messages.
        """
        return self._inspect_speak


class ClientOptionsFromEnv(
    DeepgramClientOptions
):  # pylint: disable=too-many-branches, too-many-statements
    """
    This class extends DeepgramClientOptions and will attempt to use environment variables first before defaults.
    """

    _logger: verboselogs.VerboseLogger

    def __init__(
        self,
        api_key: str = "",
        url: str = "",
        verbose: int = verboselogs.WARNING,
        headers: Optional[Dict] = None,
        options: Optional[Dict] = None,
    ):  # pylint: disable=too-many-positional-arguments
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(verboselogs.WARNING)  # temporary set for setup

        if api_key is None:
            api_key = ""

        if api_key == "":
            api_key = os.getenv("DEEPGRAM_API_KEY", "")
            if api_key == "":
                self._logger.critical("Deepgram API KEY is not set")
                raise DeepgramApiKeyError("Deepgram API KEY is not set")

        if url == "":
            url = os.getenv("DEEPGRAM_HOST", "api.deepgram.com")
            self._logger.notice(f"Deepgram host is set to {url}")

        if verbose == verboselogs.WARNING:
            _loglevel = os.getenv("DEEPGRAM_LOGGING", "")
            if _loglevel != "":
                verbose = int(_loglevel)
            if isinstance(verbose, str):
                match verbose:
                    case "NOTSET":
                        self._logger.notice("Logging level is set to NOTSET")
                        verbose = verboselogs.NOTSET
                    case "SPAM":
                        self._logger.notice("Logging level is set to SPAM")
                        verbose = verboselogs.SPAM
                    case "DEBUG":
                        self._logger.notice("Logging level is set to DEBUG")
                        verbose = verboselogs.DEBUG
                    case "VERBOSE":
                        self._logger.notice("Logging level is set to VERBOSE")
                        verbose = verboselogs.VERBOSE
                    case "NOTICE":
                        self._logger.notice("Logging level is set to NOTICE")
                        verbose = verboselogs.NOTICE
                    case "WARNING":
                        self._logger.notice("Logging level is set to WARNING")
                        verbose = verboselogs.WARNING
                    case "SUCCESS":
                        self._logger.notice("Logging level is set to SUCCESS")
                        verbose = verboselogs.SUCCESS
                    case "ERROR":
                        self._logger.notice("Logging level is set to ERROR")
                        verbose = verboselogs.ERROR
                    case "CRITICAL":
                        self._logger.notice("Logging level is set to CRITICAL")
                        verbose = verboselogs.CRITICAL
                    case _:
                        self._logger.notice("Logging level is set to WARNING")
                        verbose = verboselogs.WARNING
        self._logger.notice(f"Logging level is set to {verbose}")

        if headers is None:
            headers = {}
            for x in range(0, 20):
                header = os.getenv(f"DEEPGRAM_HEADER_{x}", None)
                if header is not None:
                    headers[header] = os.getenv(f"DEEPGRAM_HEADER_VALUE_{x}", None)
                    self._logger.debug(
                        "Deepgram header %s is set with value %s",
                        header,
                        headers[header],
                    )
                else:
                    break
            if len(headers) == 0:
                self._logger.notice("Deepgram headers are not set")
                headers = None

        if options is None:
            options = {}
            for x in range(0, 20):
                param = os.getenv(f"DEEPGRAM_PARAM_{x}", None)
                if param is not None:
                    options[param] = os.getenv(f"DEEPGRAM_PARAM_VALUE_{x}", None)
                    self._logger.debug(
                        "Deepgram option %s is set with value %s", param, options[param]
                    )
                else:
                    break
            if len(options) == 0:
                self._logger.notice("Deepgram options are not set")
                options = None

        super().__init__(
            api_key=api_key, url=url, verbose=verbose, headers=headers, options=options
        )



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import SpeakRESTClient
from .async_client import AsyncSpeakRESTClient
from .response import SpeakRESTResponse
from .options import (
    #### top level
    SpeakRESTOptions,
    SpeakOptions,
    # common
    TextSource,
    BufferSource,
    StreamSource,
    FileSource,
    # unique
    SpeakSource,
    SpeakRestSource,
    SpeakRESTSource,
)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from io import BufferedReader
from typing import Union
from typing_extensions import TypedDict


class StreamSource(TypedDict):
    """
    Represents a data source for reading binary data from a stream-like source.

    This class is used to specify a source of binary data that can be read from
    a stream, such as an audio file in .wav format.

    Attributes:
        stream (BufferedReader): A BufferedReader object for reading binary data.
    """

    stream: BufferedReader


class UrlSource(TypedDict):
    """
    Represents a data source for specifying the location of a file via a URL.

    This class is used to specify a hosted file URL, typically pointing to an
    externally hosted file, such as an audio file hosted on a server or the internet.

    Attributes:
        url (str): The URL pointing to the hosted file.
    """

    url: str


class BufferSource(TypedDict):
    """
    Represents a data source for handling raw binary data.

    This class is used to specify raw binary data, such as audio data in its
    binary form, which can be captured from a microphone or generated synthetically.

    Attributes:
        buffer (bytes): The binary data.
    """

    buffer: bytes


class TextSource(TypedDict):
    """
    Represents a data source for reading binary data from a text-like source.

    This class is used to specify a source of text data that can be read from.

    Attributes:
        text (str): A string for reading text data.
    """

    text: str


FileSource = Union[TextSource, BufferSource, StreamSource]



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import AnalyzeClient as AnalyzeClientLatest
from .v1.async_client import AsyncAnalyzeClient as AsyncAnalyzeClientLatest
from .v1.options import (
    # common
    AnalyzeOptions as AnalyzeOptionsLatest,
    UrlSource as UrlSourceLatest,
    TextSource as TextSourceLatest,
    BufferSource as BufferSourceLatest,
    StreamSource as StreamSourceLatest,
    FileSource as FileSourceLatest,
    # unique
    AnalyzeStreamSource as AnalyzeStreamSourceLatest,
    AnalyzeSource as AnalyzeSourceLatest,
)
from .v1.response import (
    SyncAnalyzeResponse as SyncAnalyzeResponseLatest,
    AnalyzeResponse as AnalyzeResponseLatest,
    AsyncAnalyzeResponse as AsyncAnalyzeResponseLatest,
    # shared
    Average as AverageLatest,
    Intent as IntentLatest,
    Intents as IntentsLatest,
    IntentsInfo as IntentsInfoLatest,
    Segment as SegmentLatest,
    SentimentInfo as SentimentInfoLatest,
    Sentiment as SentimentLatest,
    Sentiments as SentimentsLatest,
    SummaryInfo as SummaryInfoLatest,
    Topic as TopicLatest,
    Topics as TopicsLatest,
    TopicsInfo as TopicsInfoLatest,
    # unique
    Results as ResultsLatest,
    Metadata as MetadataLatest,
    Summary as SummaryLatest,
)

# The client.py points to the current supported version in the SDK.
# Older versions are supported in the SDK for backwards compatibility.

# common
UrlSource = UrlSourceLatest
TextSource = TextSourceLatest
BufferSource = BufferSourceLatest
StreamSource = StreamSourceLatest
FileSource = FileSourceLatest

AnalyzeStreamSource = AnalyzeStreamSourceLatest
AnalyzeSource = AnalyzeSourceLatest

# input
AnalyzeOptions = AnalyzeOptionsLatest

# responses
SyncAnalyzeResponse = SyncAnalyzeResponseLatest
AnalyzeResponse = AnalyzeResponseLatest
AsyncAnalyzeResponse = AsyncAnalyzeResponseLatest
# shared
Average = AverageLatest
Intent = IntentLatest
Intents = IntentsLatest
IntentsInfo = IntentsInfoLatest
Segment = SegmentLatest
SentimentInfo = SentimentInfoLatest
Sentiment = SentimentLatest
Sentiments = SentimentsLatest
SummaryInfo = SummaryInfoLatest
Topic = TopicLatest
Topics = TopicsLatest
TopicsInfo = TopicsInfoLatest
# unique
AnalyzeResults = ResultsLatest
AnalyzeMetadata = MetadataLatest
AnalyzeSummary = SummaryLatest

# clients
AnalyzeClient = AnalyzeClientLatest
AsyncAnalyzeClient = AsyncAnalyzeClientLatest


# aliases
ReadClient = AnalyzeClientLatest
AsyncReadClient = AsyncAnalyzeClientLatest



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import List, Optional

from dataclasses import dataclass, field
from dataclasses_json import config as dataclass_config

from ...common import BaseResponse

# Input


@dataclass
class ProjectOptions(BaseResponse):
    """
    Project Options
    """

    name: str = ""


@dataclass
class ModelOptions(BaseResponse):
    """
    Model Options
    """

    include_outdated: bool = False


@dataclass
class KeyOptions(BaseResponse):
    """
    Key Options
    """

    comment: Optional[str] = ""
    expiration_date: Optional[str] = field(
        default="", metadata=dataclass_config(exclude=lambda f: f == "")
    )
    time_to_live_in_seconds: Optional[int] = field(
        default=-1, metadata=dataclass_config(exclude=lambda f: f == -1)
    )
    scopes: List[str] = field(default_factory=list)
    tags: Optional[List[str]] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if _dict["scopes"] is not None:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        if _dict["tags"] is not None:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        return _dict[key]


@dataclass
class ScopeOptions(BaseResponse):
    """
    Scope Options
    """

    scope: str = ""


@dataclass
class InviteOptions(BaseResponse):
    """
    Invite Options
    """

    email: str = ""
    scope: str = ""


@dataclass
class UsageRequestOptions(BaseResponse):
    """
    Usage Request Options
    """

    start: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    end: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    limit: Optional[int] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    status: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )


@dataclass
class UsageSummaryOptions(BaseResponse):  # pylint: disable=too-many-instance-attributes
    """
    Usage Summary Options
    """

    start: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    end: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    accessor: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    tag: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    method: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    model: Optional[str] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    multichannel: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    interim_results: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    punctuate: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    ner: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    utterances: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    replace: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    profanity_filter: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    keywords: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    detect_topics: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    diarize: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    search: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    redact: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    alternatives: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    numerals: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )
    smart_format: Optional[bool] = field(
        default=None, metadata=dataclass_config(exclude=lambda f: f is None)
    )


@dataclass
class UsageFieldsOptions(BaseResponse):
    """
    Usage Fields Options
    """

    start: Optional[str] = ""
    end: Optional[str] = ""



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import AnalyzeClient
from .async_client import AsyncAnalyzeClient

# common
from .options import (
    UrlSource,
    TextSource,
    BufferSource,
    StreamSource,
    FileSource,
)

# analyze

from .options import (
    AnalyzeOptions,
    AnalyzeStreamSource,
    AnalyzeSource,
)

from .response import (
    AsyncAnalyzeResponse,
    SyncAnalyzeResponse,
    AnalyzeResponse,
    # shared
    Average,
    Intent,
    Intents,
    IntentsInfo,
    Segment,
    SentimentInfo,
    Sentiment,
    Sentiments,
    SummaryInfo,
    Topic,
    Topics,
    TopicsInfo,
    # unique
    Metadata,
    Results,
    Summary,
)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from importlib import import_module
import logging
import deprecation  # type: ignore

from .. import __version__
from .speak.v1.rest.client import SpeakRESTClient
from ..utils import verboselogs
from ..options import DeepgramClientOptions
from .errors import DeepgramModuleError


class SpeakRouter:
    """
    This class provides a Speak Clients for making requests to the Deepgram API with various configuration options.

    Attributes:
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Methods:
        rest: (Preferred) Returns a Threaded REST Client instance for interacting with Deepgram's transcription services.
        websocket: (Preferred) Returns an Threaded WebSocket Client instance for interacting with Deepgram's prerecorded transcription services.

        asyncrest: Returns an Async REST Client instance for interacting with Deepgram's transcription services.
        asyncwebsocket: Returns an Async WebSocket Client instance for interacting with Deepgram's prerecorded transcription services.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions

    def __init__(self, config: DeepgramClientOptions):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)
        self._config = config

    # when this is removed, remove --disable=W0622 from Makefile
    # pylint: disable=unused-argument
    @deprecation.deprecated(
        deprecated_in="3.4.0",
        removed_in="4.0.0",
        current_version=__version__,
        details="deepgram.speak.v1 is deprecated. Use deepgram.speak.rest or deepgram.speak.websocket instead.",
    )
    def v(self, version: str = ""):
        """
        DEPRECATED: deepgram.speak.v1 is deprecated. Use deepgram.speak.rest or deepgram.speak.websocket instead.
        """
        return SpeakRESTClient(self._config)

    # pylint: enable=unused-argument

    @property
    def rest(self):
        """
        Returns a Threaded REST Client instance for interacting with Deepgram's prerecorded Text-to-Speech services.
        """
        return self.Version(self._config, "rest")

    @property
    def asyncrest(self):
        """
        Returns an Async REST Client instance for interacting with Deepgram's prerecorded Text-to-Speech services.
        """
        return self.Version(self._config, "asyncrest")

    @property
    def websocket(self):
        """
        Returns a Threaded WebSocket Client instance for interacting with Deepgram's Text-to-Speech services.
        """
        return self.Version(self._config, "websocket")

    @property
    def asyncwebsocket(self):
        """
        Returns an Async WebSocket Client instance for interacting with Deepgram's Text-to-Speech services.
        """
        return self.Version(self._config, "asyncwebsocket")

    # INTERNAL CLASSES
    class Version:
        """
        Represents a version of the Deepgram API.
        """

        _logger: verboselogs.VerboseLogger
        _config: DeepgramClientOptions
        _parent: str

        def __init__(self, config, parent: str):
            self._logger = verboselogs.VerboseLogger(__name__)
            self._logger.addHandler(logging.StreamHandler())
            self._logger.setLevel(config.verbose)
            self._config = config
            self._parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self._parent:
        #         case "live":
        #             return LiveClient(self._config)
        #         case "prerecorded":
        #             return PreRecordedClient(self._config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(self, version: str = ""):
            """
            Returns a specific version of the Deepgram API.
            """
            self._logger.debug("Version.v ENTER")
            self._logger.info("version: %s", version)
            if len(version) == 0:
                self._logger.error("version is empty")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            type = ""
            file_name = ""
            class_name = ""
            match self._parent:
                case "websocket":
                    type = "websocket"
                    file_name = "client"
                    class_name = "SpeakWebSocketClient"
                case "asyncwebsocket":
                    type = "websocket"
                    file_name = "async_client"
                    class_name = "AsyncSpeakWebSocketClient"
                case "rest":
                    type = "rest"
                    file_name = "client"
                    class_name = "SpeakRESTClient"
                case "asyncrest":
                    type = "rest"
                    file_name = "async_client"
                    class_name = "AsyncSpeakRESTClient"
                case _:
                    self._logger.error("parent unknown: %s", self._parent)
                    self._logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.speak.v{version}.{type}.{file_name}"
            self._logger.info("path: %s", path)
            self._logger.info("class_name: %s", class_name)

            # import class
            mod = import_module(path)
            if mod is None:
                self._logger.error("module path is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, class_name)
            if my_class is None:
                self._logger.error("my_class is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            my_class = my_class(self._config)
            self._logger.notice("Version.v succeeded")
            self._logger.debug("Version.v LEAVE")
            return my_class



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from importlib import import_module
import logging

from ..utils import verboselogs
from ..options import DeepgramClientOptions
from .errors import DeepgramModuleError


class AgentRouter:
    """
    Represents a client for interacting with the Deepgram API.

    This class provides a client for making requests to the Deepgram API with various configuration options.

    Attributes:
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Raises:
        DeepgramApiKeyError: If the API key is missing or invalid.

    Methods:
        read: (Preferred) Returns an Threaded AnalyzeClient instance for interacting with Deepgram's read transcription services.
        asyncread: Returns an (Async) AnalyzeClient instance for interacting with Deepgram's read transcription services.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions

    def __init__(self, config: DeepgramClientOptions):
        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)
        self._config = config

    @property
    def websocket(self):
        """
        Returns an AgentWebSocketClient instance for interacting with Deepgram's Agent API.
        """
        return self.Version(self._config, "websocket")

    @property
    def asyncwebsocket(self):
        """
        Returns an AsyncAgentWebSocketClient instance for interacting with Deepgram's Agent API.
        """
        return self.Version(self._config, "asyncwebsocket")

    # INTERNAL CLASSES
    class Version:
        """
        Represents a version of the Deepgram API.
        """

        _logger: verboselogs.VerboseLogger
        _config: DeepgramClientOptions
        _parent: str

        def __init__(self, config, parent: str):
            self._logger = verboselogs.VerboseLogger(__name__)
            self._logger.addHandler(logging.StreamHandler())
            self._logger.setLevel(config.verbose)
            self._config = config
            self._parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self._parent:
        #         case "analyze":
        #             return AnalyzeClient(self._config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(self, version: str = ""):
            """
            Returns a specific version of the Deepgram API.
            """
            self._logger.debug("Version.v ENTER")
            self._logger.info("version: %s", version)
            if len(version) == 0:
                self._logger.error("version is empty")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            parent = ""
            file_name = ""
            class_name = ""
            match self._parent:
                case "websocket":
                    parent = "websocket"
                    file_name = "client"
                    class_name = "AgentWebSocketClient"
                case "asyncwebsocket":
                    parent = "websocket"
                    file_name = "async_client"
                    class_name = "AsyncAgentWebSocketClient"
                case _:
                    self._logger.error("parent unknown: %s", self._parent)
                    self._logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.agent.v{version}.{parent}.{file_name}"
            self._logger.info("path: %s", path)
            self._logger.info("class_name: %s", class_name)

            # import class
            mod = import_module(path)
            if mod is None:
                self._logger.error("module path is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, class_name)
            if my_class is None:
                self._logger.error("my_class is None")
                self._logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            my_class = my_class(self._config)
            self._logger.notice("Version.v succeeded")
            self._logger.debug("Version.v LEAVE")
            return my_class
